"""iotcs.sample.DirectlyConnectedDeviceSample.py."""
# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

import logging
import sys
import threading
import time

from iotcs.sample.sensors import HumiditySensor
from iotcs.client import AbstractVirtualDevice
from iotcs.client.device  import DirectlyConnectedDevice

# The IOT Cloud Service device model urn for the simulated humidity sensor
HUMIDITYSENSORMODELURN = "urn:com:oracle:iot:device:humidity_sensor"
# The device model alert format urn
HUMIDITYTOOHUMIDALERT = f'{HUMIDITYSENSORMODELURN}:too_humid'

# Keyboard watcher flag
EXITSAMPLE = False

def main():
    """
    This sample presents a simple sensor as virtual device to the IoT server.

    The simple sensor is polled every half second and each
    virtual device is updated.

    The sample runs in two modes: with policies or without policies. By
    default, the sample runs without policies. To enable the use of
    policies, set the command line option *-P*.

    When running with policies, the device policies must be uploaded to
    the server. The sample reads the sensor data and calls the
    :meth:`.VirtualDevice.offer` method, which applies the policies.
    Data and alert messages are sent
    to the server depending on the configuration of the device policies.

    When running without policies, the sample reads the sensor data and
    calls the :meth:`.VirtualDevice.set` method, which results in a
    data message being sent to the server. The sample itself must
    generate alerts if the value from the sensor exceeds
    the maximum threshold of the sensor's data model.
    """
    # Uncomment 'import logging' and the following lines to enable logging.
    #logging.basicConfig(level=logging.DEBUG)
    # Don't log urllib3 messages
    #urllib3_logger = logging.getLogger('urllib3')
    #urllib3_logger.setLevel(logging.CRITICAL)
    #urllib3_logger.setLevel(logging.ERROR)

    def usage():
        print(f'Usage: python DirectlyConnectedSample.py '
              f'[-P] <provisioner file> <passphrase>\n'
              f'where -P indicates that device policies are in effect.')

    def getFormattedTime():
        return time.strftime("%a %b %d %H:%M:%S %Z %Y")

    class HumiditySensorChangeCallback(AbstractVirtualDevice.ChangeCallback):
        """
        The sample's implementation of :class:`.AbstractVirtualDevice.ChangeCallback`. # noqa
        """
        __slots__ = (
            'sensor'
        )

        def __init__(self, sensor):
            self.sensor = sensor

        def onChange(self, event):
            """
            Called when an editable attribute is changed.

            In this sample there is only one editable attribute `maxThreshold`.
            Iterate over the list of attribute changes and
            store the new value locally in `maxHumidityThreshold`.
            :param event: :class:`.AbstractVirtualDevice.ChangeEvent` contains
            the changed attributes.

            """
            endpointId = event.getVirtualDevice().getEndpointId()
            namedValue = event.getNamedValue()
            next = namedValue
            while next is not None:
                name = next.getName()
                value = next.getValue()
                if name == "maxThreshold":
                    self.sensor.setMaxThreshold(value)
                    print(f'{getFormattedTime()} : '
                          f'{endpointId} : OnChange : {name} = {value}')
                next = next.next()

    class SampleErrorCallback(AbstractVirtualDevice.ErrorCallback):
        def onError(self, event):
            """
            An implementation of :class:`.AbstractVirtualDevice.ErrorCallback`.

            :param event: instance of VirtualDeviceBase.ErrorEvent
            :return:
            """
            virtualDevice = event.getVirtualDevice()
            displaymessage = f'{getFormattedTime()} : '\
                             f'{virtualDevice.getEndpointId()} : OnError :'
            namedValue = event.getNamedValue()
            next = namedValue
            comma = ""
            while next is not None:
                name = next.getName()
                value = next.getValue()
                displaymessage = f'{displaymessage} "{name}" = {value}{comma}'
                next = next.next()
                comma = ","

            displaymessage = f'{displaymessage} : {event.getMessage()}'
            print(f'{displaymessage}')

    # TODO: make polling interval an application argument
    # The following calculations of number_of_loops and sleepTime break the
    # SENSOR_POLLING_INTERVAL into a number of smaller intervals approximately
    # SLEEP_TIME milliseconds long. This just makes the sample a little more
    # esponsive to keyboard input and has nothing to do with the client
    # library.
    SLEEPTIME = 1000
    # sensor polling interval before sending next readings.
    SENSORPOLLINGINTERVAL = 5000
    # default is 50 loops
    NUMBEROFLOOPS = 1 if SLEEPTIME > SENSORPOLLINGINTERVAL else \
        SENSORPOLLINGINTERVAL / SLEEPTIME
    # default ends up 1 second sleep time
    sleeptime = SENSORPOLLINGINTERVAL \
        if SLEEPTIME > SENSORPOLLINGINTERVAL else \
        SLEEPTIME + (SENSORPOLLINGINTERVAL -
                     NUMBEROFLOOPS * SLEEPTIME) / NUMBEROFLOOPS
    # make sleeptime into seconds
    sleeptime = int(sleeptime/1000)

    # Set to 'True` if the *-P* command line option is set.
    usepolicies = False

    # Use the same error callbacl for both virtual devices
    errorCallback = SampleErrorCallback()
    path = None
    passphrase = None
    no_args = len(sys.argv)

    # If no_args == 4 then it is assumed that argv1 is *-P*
    if no_args == 4:
        if "-P" == sys.argv[1]:
            usepolicies = True
        else:
            usage()
            exit(1)
        path = sys.argv[2]
        passphrase = sys.argv[3]
    elif no_args < 3:
        usage()
        exit(1)
    else:
        path = sys.argv[1]
        passphrase = sys.argv[2]

    # Initialize the device client
    directlyConnectedDevice = DirectlyConnectedDevice(path, passphrase)

    # Activate the device
    if not directlyConnectedDevice.isActivated():
        print("Activating...")
        directlyConnectedDevice.activate([HUMIDITYSENSORMODELURN])

    # The "real" device
    humiditySensor = HumiditySensor(
        f'{directlyConnectedDevice.getEndpointId()}_Sample_HS')

    # Create a virtual device implementing the device model
    deviceModel = \
        directlyConnectedDevice.getDeviceModel(HUMIDITYSENSORMODELURN)

    virtualHumiditySensor = directlyConnectedDevice.createVirtualDevice(
        directlyConnectedDevice.getEndpointId(), deviceModel)

    # Initialize the sensor to the device model's maxThreshold default value
    defaultThreshold = virtualHumiditySensor.get("maxThreshold")
    humiditySensor.setMaxThreshold(defaultThreshold)

    # Monitor the virtual device for requested attribute changes and
    # errors.
    #
    # Since there is only one attribute, maxThreshold, this could have
    # done with using an attribute specific on change handler.
    #
    virtualHumiditySensor.setOnChange(
        HumiditySensorChangeCallback(humiditySensor))
    virtualHumiditySensor.setOnError(errorCallback)

    print(f'Created virtual humidity sensor '
          f'{virtualHumiditySensor.getEndpointId()}')

    tooHumidAlert = None
    if not usepolicies:
        # Create an Alert for max humidity threshold
        # Alerts do not have to be created new each time an alerting event occurs.
        tooHumidAlert = virtualHumiditySensor.createAlert(HUMIDITYTOOHUMIDALERT)

    # Flags to ensure an alert is only generated when the threshold is crossed.
    # Used without policies
    humidityAlerted = False

    # use list to manage sensor state
    # used without policies
    humiditySensorState = [tooHumidAlert, humidityAlerted]

    def processHumiditySensorDataWithoutPolicies(virtualHumidityDevice,
                                                 humiditySensor,
                                                 sensorState):
        """
        When running the sample without policies, the sample must do more work.

        The sample has to generate alerts if the sensor value exceeds the
        threshold. This logic can be handled by the client library with the
        right set of policies in place. Compare this method to that of the
        WithPolicies inner class.

        :param virtualHumidityDevice: the virtual humidity sensor
        :param humiditySensor: the real humidity sensor
        :param sensorState: track the alert status of the sensor
        :return:
        """
        humidity = humiditySensor.getHumidity()

        tooHumidAlert = sensorState[0]
        humidityAlerted = sensorState[1]

        consoleMessage = \
            f'{getFormattedTime()} : {virtualHumiditySensor.getEndpointId()} '\
            f': Set : "humidity"={humidity}'

        # Set the virtualilzed humidity sensor value in an 'update'. If
        # the humidity maxThreshold has changed, that maxThreshold attribute
        # will be added to the update (see the block below). A call to
        # 'finish' will commit the update.
        virtualHumiditySensor.update()
        virtualHumiditySensor.set("humidity", humidity)

        print(consoleMessage)
        virtualHumiditySensor.finish()

        humidityThreshold = humiditySensor.getMaxThreshold()
        if humidity > humidityThreshold:
            if not humidityAlerted:
                tooHumidAlert.set("humidity", humidity)
                print(f'{getFormattedTime()} : '
                      f'{virtualHumidityDevice.getEndpointId()} : Alert : '
                      f'Humidity="{humidity}"')
                tooHumidAlert.raise_()
                humidityAlerted = True
        else:
            humidityAlerted = False

        sensorState[0] = tooHumidAlert
        sensorState[1] = humidityAlerted

    def processHumiditySensorDataWithPolicies(virtualHumidityDevice,
                                              humiditySensor,
                                              sensorState):
        """
        Process humidity sensor data using policies.

        When running with policies, we only need to "offer" the
        humidity value. Policies can be set on the server to
        generate alerts, and filter bad data - which had to be
        handled by the sample code when running without policies.
        Compare this implementation of processSensorData to that
        the WithoutPolicies inner class.

        :param virtualHumidityDevice: the virtual humidity sensor
        :param humiditySensor: the real humidity sensor
        :param sensorState: ignored
        :return:
        """
        humidity = humiditySensor.getHumidity()

        consoleMessage = \
            f'{getFormattedTime()} : {virtualHumiditySensor.getEndpointId()} '\
            f': Offer : "humidity"={humidity}'
        print(consoleMessage)
        virtualHumiditySensor.update().offer("humidity", humidity).finish()

    # Set the appropriate sensor processing method
    processHumiditySensorData = processHumiditySensorDataWithoutPolicies
    if usepolicies is True:
        processHumiditySensorData = processHumiditySensorDataWithPolicies

    # start the keyboard watcher
    def waitforkeypress():
        input("\n\tPress Enter to quit.\n\n")
        global EXITSAMPLE
        EXITSAMPLE = True

    threading.Thread(
        name="iotcs.sample.DirectlyConnectedDeviceSample.waitforkeypress",
        target=waitforkeypress, daemon=True).start()

    while True:
        try:
            processHumiditySensorData(virtualHumiditySensor, humiditySensor,
                                      humiditySensorState)
            # Wait SENSOR_POLLING_INTERVAL seconds before sending the next
            # reading.
            # The SENSOR_POLLING_INTERVAL is broken into a number of smaller
            # increments to make the application more responsive to a keypress,
            # which will cause the application to exit.
            for i in range(int(NUMBEROFLOOPS)):
                time.sleep(sleeptime)
            if EXITSAMPLE:
                break
        except KeyboardInterrupt:
            break


if __name__ == "__main__":
    main()
