# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

import sys
import time

from iotcs.client.device import DirectlyConnectedDevice
from iotcs.client import AbstractVirtualDevice
from iotcs.client.device import VirtualDevice
from iotcs.sample.sensors import TemperatureSensor

def main():

    """
    The `DirectlyConnectedDeviceSample2` accepts two options
    - The IOT Cloud Service provisioning file for the directly connected device
    - The pass phrase used to encrypt the file

    This sample demonstrates a simple use of the Python Client Library Virtualization API. It simulates a
    humidity sensore based on the `urn:com:oracle:iot:device:humidity_sensor` device model, that must
    exist in the IOT Cloud Service from which the provisioning file was obtained. A simulated
    humidity sensor generates humidity values. Each value is set on the virtual device. If the
    `update().set().finish()` sequence is implemented, the messages are batched, and queued in the
    `MessageDispatcher` when the `finish()` method is called. If only the
    `set()` method is used a `DataMessage` is queued immediately.

    The humidity value is checked and if the value exceeds the virtual device's `maxThreshold` attribute
    an `Alert` is created and `raise_()` is called that sends the alert to the IOT Cloud Service. An `AlertMessage`
    is queued and sent to the IOT Cloud Service.
    """

    TEMPERATURESENSORMODELURN = "urn:com:oracle:iot:device:temperature_sensor"
    TEMPERATURETOOHOTALERT = f'{TEMPERATURESENSORMODELURN}:too_hot'
    TEMPERATURETOOCOLDALERT = f'{TEMPERATURESENSORMODELURN}:too_cold'

    def usage():
        print("Usage: python DirectlyConnectedSample2.py <provisioner file> <passphrase>")

    def getFormattedTime():
        return time.strftime("%a %b %d %H:%M:%S %Z %Y")

    # TemperatureSensor callback classes and methods
    class TemperatureSensorPowerCallable(VirtualDevice.Callable):

        __slots__ = (
            'sensor'
        )

        def __init__(self, sensor):
            self.sensor = sensor

        def call(self, virtualDevice, on):
            if on is None:
                print(f'{getFormattedTime()} : {virtualDevice.getEndpointId()} : Call : "power"=None')
                return

            self.sensor.power(on)
            print(f'{getFormattedTime()} : {virtualDevice.getEndpointId()} : Call : "power"={on}')

    class TemperatureSensorResetCallable(VirtualDevice.Callable):

        __slots__ = (
            'sensor'
        )

        def __init__(self, sensor):
            self.sensor = sensor

        def call(self, virtualDevice, reset):
            # reset is ignored
            self.sensor.reset()
            print(f'{getFormattedTime()} : {virtualDevice.getEndpointId()} : Call : reset')

    class TemperatureSensorChangeCallback(AbstractVirtualDevice.ChangeCallback):
        """
        The sample's implementation of `AbstractVirtualDevice.ChangeCallback` called when an
        editable `VirtualDevice` attribute is changed.
        """
        __slots__ = (
            'sensor'
        )

        def __init__(self, sensor):
            self.sensor = sensor

        def onChange(self, event):
            """
            `event` contains the list of changed attributes. In temperature sensor model there
            are two editable attributes, `minThreshold` and `maxThreshold`. Iterate over the
            list of attribute changes and store the new value in the sensor`.
            :param event: `AbstractVirtualDevice.ChangeEvent` contains the changed attributes.
            """
            namedValue = event.getNamedValue()
            next = namedValue
            while next is not None:
                name = next.getName()
                value = next.getValue()
                if name == "maxThreshold":
                    self.sensor.setMaxThreshold(value)
                    print(f'{getFormattedTime()} : {dcd.getEndpointId()} : OnChange : {name} = {value}')
                elif name == "minThreshold":
                    self.sensor.setMinThreshold(value)
                    print(f'{getFormattedTime()} : {dcd.getEndpointId()} : OnChange : {name} = {value}')

                next = next.next()

    class SampleErrorCallback(AbstractVirtualDevice.ErrorCallback):
        def onError(self, event):
            """
            :param event: instance of VirtualDeviceBase.ErrorEvent
            :return:
            """
            virtualDevice = event.getVirtualDevice()
            displaymessage = f'{getFormattedTime()} : {virtualDevice.getEndpointId()}'
            namedValue = event.getNamedValue()
            next = namedValue
            while next is not None:
                name = next.getName()
                value = next.getValue()
                displaymessage = f'{displaymessage}\n\t{name} = {value}'
                next = next.next()

            displaymessage = f'{displaymessage}\nerror message: {event.getMessage()}'
            print(f'{displaymessage}')

    path = None
    passphrase = None
    no_args = len(sys.argv)

    if no_args < 3:
        usage()
        exit(1)
    else:
        path = sys.argv[1]
        passphrase = sys.argv[2]

    # Create directly connected device (DCD)
    dcd = DirectlyConnectedDevice(path, passphrase)

    # Check if device is not activated
    if not dcd.isActivated():
        # Activate the DCD
        print("Activating...")
        dcd.activate([TEMPERATURESENSORMODELURN])

    # Get the device model from IoT CS
    temperatureDM = dcd.getDeviceModel(TEMPERATURESENSORMODELURN)

    # Create the Temperature sensor object
    temperatureSensor = TemperatureSensor(dcd.getEndpointId())

    # Create the virtual device for DCD
    virtualTemperatureSensor = dcd.createVirtualDevice(dcd.getEndpointId(), temperatureDM)

    # TODO: set independent attribute change handlers.
    virtualTemperatureSensor.setOnChange(TemperatureSensorChangeCallback(temperatureSensor))
    virtualTemperatureSensor.setCallable("power", TemperatureSensorPowerCallable(temperatureSensor))
    virtualTemperatureSensor.setCallable("reset", TemperatureSensorResetCallable(temperatureSensor))
    virtualTemperatureSensor.setOnError(SampleErrorCallback())

    maxTemperatureThreshold = virtualTemperatureSensor.get("maxThreshold")
    temperatureSensor.setMaxThreshold(maxTemperatureThreshold)

    minTemperatureThreshold = virtualTemperatureSensor.get("minThreshold")
    temperatureSensor.setMinThreshold(minTemperatureThreshold)

    # Create the too_hot alert
    tooHotAlert = virtualTemperatureSensor.createAlert(TEMPERATURETOOHOTALERT)
    tooColdAlert = virtualTemperatureSensor.createAlert(TEMPERATURETOOCOLDALERT)

    print("Press Enter to Exit")

    tempAlerted = False
    wasOff = False
    prevMinTemp = float("inf")
    prevMaxTemp = float("-inf")

    try:
        while True:

            if not temperatureSensor.isPoweredOn():
                time.sleep(5)
                continue

            # Read the latest sensor temperature
            temperature = float(temperatureSensor.getTemp())

            # Data model has temp range as [-20,80]. If data is out
            # of range, do not set the attribute in the virtual device.
            # Note: without this check, the set would raise ValueError
            # update the virtual device with latest data from sensor
            if -20 >= temperature or temperature >= 80:
                time.sleep(5)
                continue

            # Open an update transaction
            virtualTemperatureSensor.update()

            # update the virtual device with latest data from sensor
            consoleMessage = f'{getFormattedTime()} : {virtualTemperatureSensor.getEndpointId()}: Set : "temp"= {temperature}'
            virtualTemperatureSensor.set("temp", float(temperature))

            unit = temperatureSensor.getUnit()
            minTemp = temperatureSensor.getMinTemp()
            maxTemp = temperatureSensor.getMaxTemp()

            if wasOff:
                tempAlerted = False
                wasOff = False
                restartTime = temperatureSensor.getStartTime()
                consoleMessage = f'{consoleMessage},"unit"={unit}, "startTime"={restartTime}'
                virtualTemperatureSensor.set("startTime", restartTime)
                virtualTemperatureSensor.set("unit", unit)

            if minTemp != prevMinTemp:
                prevMinTemp = minTemp
                consoleMessage = f'{consoleMessage},"minTemp"={minTemp}'
                virtualTemperatureSensor.set("minTemp", float(minTemp))

            if maxTemp != prevMaxTemp:
                prevMaxTemp = maxTemp
                consoleMessage = f'{consoleMessage},"maxTemp"={maxTemp}'
                virtualTemperatureSensor.set("maxTemp", float(maxTemp))

            # send the data message to IoT CS
            print(consoleMessage)
            virtualTemperatureSensor.finish()

            if temperature > temperatureSensor.getMaxThreshold():
                if not tempAlerted:
                    consoleMessage = f'{getFormattedTime()} : {virtualTemperatureSensor.getEndpointId()} : TooHotAlert : "temp"={temperature}'
                    tooHotAlert.set("temp", temperature)
                    tooHotAlert.set("unit", "C")
                    tooHotAlert.set("maxThreshold", temperatureSensor.getMaxThreshold())
                    tempAlerted = True
                    print(f'{consoleMessage} : "maxThreshold"={temperatureSensor.getMaxThreshold()}')
                    tooHotAlert.raise_()
            elif temperature < temperatureSensor.getMinThreshold():
                if not tempAlerted:
                    consoleMessage = f'{getFormattedTime()} : {virtualTemperatureSensor.getEndpointId()} : TooColdAlert : "temp"={temperature}'
                    tooColdAlert.set("temp", temperature)
                    tooColdAlert.set("unit", "C")
                    tooColdAlert.set("minThreshold", temperatureSensor.getMinThreshold())
                    tempAlerted = True
                    print(f'{consoleMessage} : "minThreshold"={temperatureSensor.getMinThreshold()}')
                    tooColdAlert.raise_()
            else:
                tempAlerted = False

            # Sleep for 5 seconds
            time.sleep(5)
    except KeyboardInterrupt:
        pass

    dcd.close()


if __name__ == "__main__":
    main()
