# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

import sys
import time
import datetime
from random import randint

from iotcs.client.device import DirectlyConnectedDevice

no_args = len(sys.argv)
maxTemperatureThreshold = 80
virtualDevice = None
tooHumidAlert = None
dcd = None
isPoweredOn = True

def usage():
    print("Usage:");
    print("\t python DirectlyConnectedSample.py <provisioner file> <passphrase>")

def getFormattedTime():
    return time.strftime("%a %b %d %H:%M:%S %Z %Y")

def checkAlert(humidity):
    # Check for threshold and raise alert
    global tooHumidAlert, virtualDevice
    if (humidity > maxTemperatureThreshold):
        tooHumidAlert.humidity = humidity
        print(getFormattedTime()  + " : " + dcd.getEndpointId() +
              " : Alert : Humidity='" + str(humidity) + "'")
        tooHumidAlert.Raise()

def onChangeCallbackHandler(changedAttributes:dict()):
    global maxTemperatureThreshold
    for (name, value) in changedAttributes.items():
        if (name == "maxThreshold"):
            maxThreshold = value
        print("{} : {} : OnChange : {} = {}".format(getFormattedTime(),
                                                    dcd.getEndpointId(), name, str(value)))

def OnErrorCallBack(msgList):
    print("Error callback")

def main():
    if (no_args < 3):
        usage()
        exit(1)
    else:
        path = sys.argv[1]
        passphrase   = sys.argv[2]

    global dcd,virtualDevice,maxTemperatureThreshold, tooHumidAlert, isPoweredOn
    # Create directly connected device (DCD)
    dcd = DirectlyConnectedDevice(path, passphrase)
    humidityURN = "urn:com:oracle:iot:device:test_humidity_sensor"
    tooHumidAlertURN = humidityURN + ":too_humid"
    pressureDataURN = humidityURN + ":pressureData"

    # Check if device is not activated
    if (not dcd.isActivated()):
        # Activate the DCD
        dcd.activate([humidityURN])

    # Get the device model from IoT CS and parse it.
    humidityDM = dcd.getDeviceModel(humidityURN)

    # Get the Humidity sensor object
    sensor = Sensor()

    # Create the virtual device for DCD
    virtualDevice = dcd.createVirtualDevice(dcd.getEndpointId(), humidityDM)

    virtualDevice.setOnChangeCallback(onChangeCallbackHandler)

    maxThreshold = virtualDevice.maxThreshold

    # Create the alert.
    tooHumidAlert = virtualDevice.createAlert(tooHumidAlertURN)
    # Create the custom data
    pressureData = virtualDevice.createData(pressureDataURN)

    print("Press Enter to Exit")
    try:
        while True:
            if isPoweredOn:
                humidity = sensor.getHumidity()
                pressure = sensor.getPressure()

                # update the virtual device with latest data from sensor
                virtualDevice.humidity = humidity
                print(getFormattedTime()  + " : " + dcd.getEndpointId() +
                      " : Set : Humidity='" + str(pressure) + "'")

                pressureData.pressure = pressure
                pressureData.time = int(datetime.datetime.now().strftime("%s")) * 1000
                pressureData.submit()

                print(getFormattedTime()  + " : " + dcd.getEndpointId() +
                      " : Data : Pressure='" + str(pressure) + "'")

                checkAlert(humidity)

            # Sleep for 5 seconds
            time.sleep(5)
    except KeyboardInterrupt:
        pass

    dcd.close()

class Sensor(object):
    MIN_HUMIDITY = 10
    MAX_HUMIDITY = 100

    MIN_PRESSURE = 20
    MAX_PRESSURE = 60

    def __init__(self, threshold = MAX_HUMIDITY):
        self.maxThreshold = threshold

    def getHumidity(self):
        return randint(Sensor.MIN_HUMIDITY, Sensor.MAX_HUMIDITY)

    def getPressure(self):
        return randint(Sensor.MIN_PRESSURE, Sensor.MAX_PRESSURE)


if __name__ == "__main__":
    main()
