# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

import logging
import sys
import threading
import time

from iotcs.client import AbstractVirtualDevice
from iotcs.client.device import GatewayDevice
from iotcs.client.device import VirtualDevice

from iotcs.sample.sensors import HumiditySensor
from iotcs.sample.sensors import TemperatureSensor

from iotcs.common import Util

TEMPERATURESENSORMODELURN = "urn:com:oracle:iot:device:temperature_sensor"
TEMPERATURETOOHOTALERT = f'{TEMPERATURESENSORMODELURN}:too_hot'
TEMPERATURETOOCOLDALERT = f'{TEMPERATURESENSORMODELURN}:too_cold'

HUMIDITYSENSORMODELURN = "urn:com:oracle:iot:device:humidity_sensor"
HUMIDITYTOOHUMIDALERT = f'{HUMIDITYSENSORMODELURN}:too_humid'

# Keyboard watcher flag
EXITSAMPLE = False

def main():
    """
    This sample is a gateway that presents multiple simple sensors as virtual
    devices to the IoT server. The simple sensors are polled every half second
    and each virtual device is updated.

    The sample runs in two modes: with policies or without policies. By
    default, the sample runs without policies. To enable the use of
    policies, set the command line option *-P*.

    When running with policies, the device policies must be uploaded to
    the server. The sample reads the humidity and temperature values
    from the sensors and calls the :meth:`.VirtualDevice.offer` method,
    which applies the policies. Data and alert messages are sent
    to the server depending on the configuration of the device policies.

    When running without policies, the sample reads the sensor data and
    calls the :meth:`.VirtualDevice.set` method, which results in a
    data message being sent to the server. The sample itself must
    generate alerts if the value from either sensor exceeds
    the maximum threshold of that sensor's data model.
    """

    # Uncomment the following line to enable logging.
    #logging.basicConfig(level=logging.DEBUG)
    # Don't log urllib3 messages
    #urllib3_logger = logging.getLogger('urllib3')
    #urllib3_logger.setLevel(logging.CRITICAL)

    def usage():
        print(f'Usage: python DirectlyConnectedSample.py '
              f'[-P] <provisioner file> <passphrase> '
              f'[temperatureICD ID] [humidityICD ID]\n'
              f'where -P indicates that device policies are in effect\n'
              f'A temperatureICD ID must be specified if a '
              f'humidityICD ID is specified')

    def getFormattedTime(ts=None):
        """
        Return `strftime` formatted time for `ts`.

        :param ts: in milliseconds
        :return: formatted time
        """
        if ts is None:
            return time.strftime("%a %b %d %H:%M:%S %Z %Y")
        else:
            return time.strftime("%a %b %d %H:%M:%S %Z %Y", time.localtime(ts/1000.0))

    class TemperatureSensorPowerActionCallback(VirtualDevice.ActionCallback):
        """
        The temperatureSensor model has a *power* action.

        The action takes a boolean argument: `True` to power on the simulated device,
        `False` to power off the simulated device.
        """
        __slots__ = (
            'sensor'
        )

        def __init__(self, sensor):
            self.sensor = sensor

        def onAction(self, actionEvent):
            virtualDevice = actionEvent.getVirtualDevice()
            arg = actionEvent.getNamedValue()
            on = arg.getValue()
            consoleMessage = f'{getFormattedTime()} : '\
                f'{virtualDevice.getEndpointId()} : onAction : "power"='
            if on is None:
                on = False
            elif isinstance(on, (str,)):
                on = Util.strtobool(on)
            elif isinstance(on, (bool,)):
                on = on

            isOn = self.sensor.isPoweredOn()
            if on != isOn:
                self.sensor.power(on)

            print(f'{consoleMessage}{on}')

    class TemperatureSensorResetActionCallback(VirtualDevice.ActionCallback):
        """
        The temperatureSensor model has a *reset* action.

        The *reset* action resets the temperature sensor to factory defaults.
        """
        __slots__ = (
            'sensor'
        )

        def __init__(self, sensor):
            self.sensor = sensor

        def onAction(self, actionEvent):
            virtualDevice = actionEvent.getVirtualDevice()
            consoleMessage = f'{getFormattedTime()} : '\
                  f'{virtualDevice.getEndpointId()} : onAction : reset'
            if self.sensor.isPoweredOn():
                self.sensor.reset()
            else:
                msg = f'{consoleMessage} ignored. Device is powered off'

            print(consoleMessage)

    class TemperatureSensorChangeCallback(AbstractVirtualDevice.ChangeCallback):
        """
        An implementation of :class:`.AbstractVirtualDevice.ChangeCallback`.
        """
        __slots__ = (
            'sensor'
        )

        def __init__(self, sensor):
            self.sensor = sensor

        def onChange(self, event):
            """
            Called when an editable attribute is changed.

            The temperature sensor model has two writable attributes,
            *minThreshold* and *maxThreshold*. Iterate over the
            list of attribute changes and store the new value in the sensor`.
            :param event: :class:`.AbstractVirtualDevice.ChangeEvent` contains
                the changed attributes.
            """
            namedValue = event.getNamedValue()
            next = namedValue
            virtualDevice = event.getVirtualDevice()
            consoleMessage = f'{getFormattedTime()} : ' \
                         f'{virtualDevice.getEndpointId()} : onChange : '
            comma = ""
            while next is not None:
                name = next.getName()
                value = next.getValue()
                if name == "maxThreshold":
                    self.sensor.setMaxThreshold(value)
                    consoleMessage=f'{consoleMessage}{comma}{name}={value}'
                elif name == "minThreshold":
                    self.sensor.setMinThreshold(value)
                    consoleMessage=f'{consoleMessage}{comma}{name}={value}'
                comma = ","
                next = next.next()

            if not self.sensor.isPoweredOn():
                consoleMessage = f'{consoleMessage} ignored. Device is powered off'

            print(consoleMessage)

    ####
    # HumiditySensor callback classes methods

    class HumiditySensorChangeCallback(AbstractVirtualDevice.ChangeCallback):
        """
        An implementation of :class:`.AbstractVirtualDevice.ChangeCallback`.
        """
        __slots__ = (
            'sensor'
        )

        def __init__(self, sensor):
            self.sensor = sensor

        def onChange(self, event):
            """
            Called when an editable attribute is changed.

            The humidity sensor model has one editable attribute
            *maxThreshold*. Iterate over the list of attribute changes and
            store the new value in the sensor.
            :param event: :class:`.AbstractVirtualDevice.ChangeEvent` contains
                the changed attributes.
            """
            namedValue = event.getNamedValue()
            next = namedValue
            virtualDevice = event.getVirtualDevice()
            consoleMessage = f'{getFormattedTime()} : ' \
                f'{virtualDevice.getEndpointId()} : onChange : '
            comma = ""
            while next is not None:
                name = next.getName()
                value = next.getValue()
                if name == "maxThreshold":
                    self.sensor.setMaxThreshold(value)
                    print(f'{consoleMessage}{comma}"{name}"={value}')
                comma = ","
                next = next.next()

    # The same error callback can support both sensors
    class SampleErrorCallback(AbstractVirtualDevice.ErrorCallback):
        def onError(self, event):
            """
            An implementation of :class:`.AbstractVirtualDevice.ErrorCallback`.

            :param event: instance of VirtualDeviceBase.ErrorEvent
            :return:
            """
            virtualDevice = event.getVirtualDevice()
            consoleMessage = f'{getFormattedTime()} : '\
                f'{virtualDevice.getEndpointId()} : onError '
            namedValue = event.getNamedValue()
            next = namedValue
            comma = ""
            while next is not None:
                name = next.getName()
                value = next.getValue()
                consoleMessage = f'{consoleMessage}{comma}"{name}"={value}'
                next = next.next()
                comma = ","

            consoleMessage = f'{consoleMessage} : {event.getMessage()}'
            print(f'{consoleMessage}')

    ####

    # TODO: make polling interval an application argument
    # The following calculations of number_of_loops and sleepTime break the
    # SENSOR_POLLING_INTERVAL into a number of smaller intervals approximately
    # SLEEP_TIME milliseconds long. This just makes the sample a little more
    # esponsive to keyboard input and has nothing to do with the client
    # library.
    SLEEPTIME = 1000
    SENSORPOLLINGINTERVAL = 5000
    # default is 5 loops
    NUMBEROFLOOPS = int(1 if SLEEPTIME > SENSORPOLLINGINTERVAL else
                        SENSORPOLLINGINTERVAL/SLEEPTIME)
    # default ends up 1 second sleep time
    sleeptime = SENSORPOLLINGINTERVAL \
        if SLEEPTIME > SENSORPOLLINGINTERVAL else \
            SLEEPTIME + (SENSORPOLLINGINTERVAL -
                     NUMBEROFLOOPS * SLEEPTIME) / NUMBEROFLOOPS
    # make sleeptime into seconds
    sleeptime = int(sleeptime/1000)

    # Set to 'True` if the *-P* command line option is set.
    usepolicies = False

    # Use the same error callbacl for both virtual devices
    errorCallback = SampleErrorCallback()
    path = None
    passphrase = None
    humiditySensorId = None
    temperatureSensorId = None
    argslen = len(sys.argv)
    nextarg = 0

    # There are 5 potential arguments
    # -P (optional)
    # <provisioning file> (required)
    # <passphrase> (required)
    # [temperature sensor device id] (optional)
    # [humidity sensor device id] (optional)
    if argslen < 3:
        # required args missing
        usage()
        exit(1)
    nextarg = 1
    if argslen > 3:
        if sys.argv[nextarg] == "-P":
            usepolicies = True
            nextarg = 2
    path = sys.argv[nextarg]
    passphrase = sys.argv[nextarg + 1]
    nextarg +=2
    # have icd ids ?
    if argslen > nextarg:
        temperatureSensorId = sys.argv[nextarg]
        nextarg += 1
    if argslen > nextarg:
        humiditySensorId = sys.argv[nextarg]

    print("Creating Gateway instance...")
    gatewayDevice = GatewayDevice(path, passphrase)

    # Actiate the device
    if not gatewayDevice.isActivated():
        print("Activating...")
        # TODO: Change arguments to activate, make "default" None
        gatewayDevice.activate([])

    # The hardware id for the sensor instances are fabricated
    # from the gateway's endpoint id and a suffix, the combination of
    # which is likely to be unique. On subsequent runs of the
    # GatewayDeviceSample (with the same gateway device), the
    # endpoint id from the previously registered indirectly connected
    # device will be returned by the registerDevice call. In other
    # words, the same indirectly connected device will be used.

    # Configure the TemperatureSensor
    # If the user gave a hardware id for the temperature sensor,
    # then restrict the sensor to this gateway. This means that
    # the sensor cannot be connected through other gateways.
    temperatureSensorRestricted = False
    if temperatureSensorId is None:
        temperatureSensorId = gatewayDevice.getEndpointId() + "_Sample_TS"
    else:
        temperatureSensorRestricted = True
    temperatureSensor = TemperatureSensor(temperatureSensorId)

    metadata = dict()
    metadata[GatewayDevice.MANUFACTURER] = temperatureSensor.getManufacturer()
    metadata[GatewayDevice.MODEL_NUMBER] = temperatureSensor.getModelNumber()
    metadata[GatewayDevice.SERIAL_NUMBER] = temperatureSensor.getSerialNumber()

    temperatureSensorEndpointId = \
        gatewayDevice.registerDevice(temperatureSensor.getHardwareId(),
                                     metadata,
                                     [TEMPERATURESENSORMODELURN],
                                     temperatureSensorRestricted)
    # Create the virtual devices implementing the device models
    temperatureDeviceModel = \
        gatewayDevice.getDeviceModel(TEMPERATURESENSORMODELURN)
    virtualTemperatureSensor = \
        gatewayDevice.createVirtualDevice(temperatureSensorEndpointId,
                                          temperatureDeviceModel)

    # Set callbacks on the virtual devices to set attributes and
    # control the actual devices.
    virtualTemperatureSensor.setOnChange(
        TemperatureSensorChangeCallback(temperatureSensor))
    # Create the callback to handle calling reset
    # the temperature sensor device.
    virtualTemperatureSensor.setOnAction(
        TemperatureSensorResetActionCallback(temperatureSensor), "reset")
    # Create the callback to handle calling power *on* and power *off* on
    # the temperature sensor device.
    virtualTemperatureSensor.setOnAction(
        TemperatureSensorPowerActionCallback(temperatureSensor), "power")
    virtualTemperatureSensor.setOnError(errorCallback)

    print(f'Created virtual temperature sensor {temperatureSensorEndpointId}')

    # Configure the HumiditySensor
    # If the user gave a hardware id for the humidity sensor,
    # then restrict the sensor to this gateway. This means that
    # the sensor cannot be connected through other gateways.
    humiditySensorRestricted = False
    if humiditySensorId is None:
        humiditySensorId = f'{gatewayDevice.getEndpointId()}_Sample_HS'
    else:
        humiditySensorRestricted = True
    humiditySensor = HumiditySensor(humiditySensorId)

    metadata = dict()
    metadata[GatewayDevice.MANUFACTURER] = humiditySensor.getManufacturer()
    metadata[GatewayDevice.MODEL_NUMBER] = humiditySensor.getModelNumber()
    metadata[GatewayDevice.SERIAL_NUMBER] = humiditySensor.getSerialNumber()

    humiditySensorEndpointId = \
        gatewayDevice.registerDevice(humiditySensor.getHardwareId(),
                                     metadata,
                                     [HUMIDITYSENSORMODELURN],
                                     humiditySensorRestricted)
    # Create the virtual devices implementing the device models
    humidityDeviceModel = gatewayDevice.getDeviceModel(HUMIDITYSENSORMODELURN)
    virtualHumiditySensor = \
        gatewayDevice.createVirtualDevice(humiditySensorEndpointId,
                                          humidityDeviceModel)

    # Initialize the humidity sensor to the device model's maxThreshold
    # default value
    defaultThreshold = virtualHumiditySensor.get("maxThreshold")
    humiditySensor.setMaxThreshold(defaultThreshold)

    # Set callbacks on the virtual devices to set attributes and
    # control the actual devices.
    virtualHumiditySensor.setOnChange(HumiditySensorChangeCallback(
        humiditySensor))
    virtualHumiditySensor.setOnError(errorCallback)

    print(f'Created virtual humidity sensor {humiditySensorEndpointId}')

    # For TemperatureSensor state changes
    wasOff = True

    tooHotAlert = None
    tooColdAlert = None
    if not usepolicies:
        # Create an Alert for max/min temperature threshold
        # Alerts do not have to be created new each time an alerting event occurs.
        tooHotAlert = virtualTemperatureSensor.createAlert(TEMPERATURETOOHOTALERT)
        tooColdAlert = \
            virtualTemperatureSensor.createAlert(TEMPERATURETOOCOLDALERT)

    # Only used without policies
    # Flags to ensure an alert is only generated when the threshold is crossed.
    tempAlerted = False
    # Only used without policies
    # The previous minimum temperature.
    # If the new minimum temperature is less than the previous,
    # update "minTemp" in the virtual TemperatureSensor
    prevMinTemp = float('inf')
    # Only used without policies
    # The previous maximum temperature.
    # If the new maximum temperature is less than the previous,
    # update "maxTemp" in the virtual TemperatureSensor
    prevMaxTemp = float('-inf')

    # Only 'wasOff' is used with policies
    # Create list to maintain state
    temperatureSensorState = [wasOff, tooHotAlert, tooColdAlert, tempAlerted,
                              prevMinTemp, prevMaxTemp]

    def processTemperatureSensorDataWithoutPolicies(virtualTemperatureSensor,
                                     temperatureSensor, sensorState):
        """
        Collect temperature and humidity data.

        Collect temperature and humidity data and set the values on the
        virtual devices. Handle alerts from the the virtual devices.
        :return: True while sensor data is collected, and False when the user
        stops the application by pressing Enter.
        """

        wasOff = sensorState[0]
        tooHotAlert = sensorState[1]
        tooColdAlert = sensorState[2]
        tempAlerted = sensorState[3]
        prevMinTemp = sensorState[4]
        prevMaxTemp = sensorState[5]

        if not temperatureSensor.isPoweredOn():
            sensorState[0] = True
            return

        # Get the latest temperature from the sensor
        temperature = temperatureSensor.getTemp()
        unit = temperatureSensor.getUnit()
        minTemp = temperatureSensor.getMinTemp()
        maxTemp = temperatureSensor.getMaxTemp()
        endpointId = virtualTemperatureSensor.getEndpointId()

        consoleMessage = f'{getFormattedTime()} : '\
                         f'{endpointId} : Set : '
        comma = ""

        # Open an update transaction
        virtualTemperatureSensor.update()

        # Data model has temp range as [-20,80]. If data is out
        # of range, do not set the attribute in the virtual device.
        # Note: without this check, the set would raise ValueError
        # update the virtual device with latest data from sensor
        if -20 < temperature and temperature < 80:
            consoleMessage = f'{consoleMessage}"temp"={temperature}'
            virtualTemperatureSensor.set("temp", float(temperature))
            comma = ","

        # There is no need to set startTime or units if the value has not changed.
        if wasOff:
            tempAlerted = False
            wasOff = False
            restartTime = temperatureSensor.getStartTime()
            formattedtime = getFormattedTime(ts=restartTime)
            consoleMessage = \
                f'{consoleMessage}{comma}"unit"={unit},"startTime"={formattedtime}'
            virtualTemperatureSensor.set("startTime", restartTime)
            virtualTemperatureSensor.set("unit", unit)
            comma = ","

        if minTemp != prevMinTemp:
            prevMinTemp = minTemp
            consoleMessage = f'{consoleMessage}{comma}"minTemp"={minTemp}'
            virtualTemperatureSensor.set("minTemp", float(minTemp))
            comma = ","

        if maxTemp != prevMaxTemp:
            prevMaxTemp = maxTemp
            consoleMessage = f'{consoleMessage}{comma}"maxTemp"={maxTemp}'
            virtualTemperatureSensor.set("maxTemp", float(maxTemp))

        # finish() commits the update.
        print(consoleMessage)
        virtualTemperatureSensor.finish()

        maxthreshold = temperatureSensor.getMaxThreshold()
        minthreshold = temperatureSensor.getMinThreshold()
        if temperature > maxthreshold:
            if not tempAlerted:
                tempAlerted = True
                consoleMessage = \
                    f'{getFormattedTime()} : {endpointId} : Alert : '\
                    f'"temp"={temperature},'\
                    f'"maxThreshold"={maxthreshold},'\
                    f'"unit"={unit}'
                print(consoleMessage)
                tooHotAlert.set("temp", float(temperature))
                tooHotAlert.set("unit", temperatureSensor.getUnit())
                tooHotAlert.set("maxThreshold", float(maxthreshold))
                tooHotAlert.raise_()
        elif temperature < minthreshold:
            if not tempAlerted:
                tempAlerted = True
                consoleMessage = \
                    f'{getFormattedTime()} : {endpointId} : Alert :'\
                    f'"temp"={temperature},' \
                    f'"minThreshold"={minthreshold},' \
                    f'"unit"={unit}'
                print(consoleMessage)
                tooColdAlert.set("temp", float(temperature))
                tooColdAlert.set("unit", temperatureSensor.getUnit())
                tooColdAlert.set("minThreshold", float(minthreshold))
                tooColdAlert.raise_()
        else:
            tempAlerted = False

        sensorState[0] = wasOff
        sensorState[1] = tooHotAlert
        sensorState[3] = tooColdAlert
        sensorState[3] = tempAlerted
        sensorState[4] = prevMinTemp
        sensorState[5] = prevMaxTemp

    def processTemperatureWithPolicies(virtualTemperatureSensor,
                                       temperatureSensor, sensorState):
        """
        Collect temperature and humidity data.

        Collect temperature and humidity data and set the values on the
        virtual devices. Handle alerts from the the virtual devices.
        :return: True while sensor data is collected, and False when the user
        stops the application by pressing Enter.
        """

        wasOff = sensorState[0]
        if not temperatureSensor.isPoweredOn():
            sensorState[0] = True
            return

        # Get the latest temperature from the sensor
        temperature = temperatureSensor.getTemp()


        consoleMessage = f'{getFormattedTime()} : ' \
            f'{virtualTemperatureSensor.getEndpointId()} : Offer : ' \
            f'"temp"={temperature}'
        # Open an update transaction
        virtualTemperatureSensor.update()\
            .offer("temp", float(temperature))

        if wasOff:
            wasOff = False
            restartTime = temperatureSensor.getStartTime()
            formattedtime = getFormattedTime(ts=restartTime)
            consoleMessage = \
                f'{consoleMessage},"startTime"={formattedtime}'
            virtualTemperatureSensor.offer("startTime", restartTime)

        # finish() commits the update.
        print(consoleMessage)
        virtualTemperatureSensor.finish()

        sensorState[0] = wasOff

    # The default is not to use policies
    processTemperatureSensorData = processTemperatureSensorDataWithoutPolicies
    if usepolicies == True:
        processTemperatureSensorData = processTemperatureWithPolicies

    # Only used without policies
    # Create an Alert for max humidity threshold
    # Alerts do not have to be created new each time an alerting event occurs.
    tooHumidAlert = None
    if not usepolicies:
        tooHumidAlert = virtualHumiditySensor.createAlert(HUMIDITYTOOHUMIDALERT)

    # Flags to ensure an alert is only generated when the threshold is crossed.
    humidityAlerted = False

    # Use list to maintain sensor state
    # used without policies
    humiditySensorState = [tooHumidAlert, humidityAlerted]

    def processHumiditySensorDataWithoutPolicies(virtualHumiditySensor,
                                                 humiditySensor,
                                                 sensorState):
        """
        When running the sample without policies, the sample must do more work.
        The sample has to generate alerts if the sensor value exceeds the
        threshold. This logic can be handled by the client library with the
        right set of policies in place. Compare this method to that of the
        WithPolicies inner class.

        :param virtualHumidityDevice: the virtual humidity sensor
        :param humiditySensor: the real humidity sensor
        :param sensorState: track the alert status of the sensor
        :return:
        """

        humidity = humiditySensor.getHumidity()

        tooHumidAlert = sensorState[0]
        humidityAlerted = sensorState[1]

        consoleMessage = \
            f'{getFormattedTime()} : {virtualHumiditySensor.getEndpointId()} '\
            f': Set : "humidity"={humidity}'

        # Open an update transaction
        virtualHumiditySensor.update()
        virtualHumiditySensor.set("humidity", humidity)

        print(consoleMessage)
        virtualHumiditySensor.finish()

        humidityThreshold = humiditySensor.getMaxThreshold()
        if humidity > humidityThreshold:
            if not humidityAlerted:
                consoleMessage = f'{getFormattedTime()} : '\
                        f'{virtualHumiditySensor.getEndpointId()} : '\
                        f'Alert : "humidity"={humidity}'
                tooHumidAlert.set("humidity", humidity)
                tooHumidAlert.raise_()
                humidityAlerted = True
                print(consoleMessage)
        else:
            humidityAlerted = False

        sensorState[0] = tooHumidAlert
        sensorState[1] = humidityAlerted

    def processHumiditySensorDataWithPolicies(virtualHumiditySensor,
                                              humiditySensor,
                                              sensorState):
        """
        Process humidity sensor data using policies.

        When running with policies, we only need to "offer" the
        humidity value. Policies can be set on the server to
        generate alerts, and filter bad data - which had to be
        handled by the sample code when running without policies.
        Compare this implementation of processSensorData to that
        the WithoutPolicies inner class.

        :param virtualHumidityDevice: the virtual humidity sensor
        :param humiditySensor: the real humidity sensor
        :param sensorState: ignored
        :return:
        """
        humidity = humiditySensor.getHumidity()

        consoleMessage = \
            f'{getFormattedTime()} : {virtualHumiditySensor.getEndpointId()} ' \
            f': Offer : "humidity"={humidity}'
        print(consoleMessage)
        virtualHumiditySensor.update().offer("humidity", humidity).finish()

    # Set the appropriate sensor processing method
    processHumiditySensorData = processHumiditySensorDataWithoutPolicies
    if usepolicies == True:
        processHumiditySensorData = processHumiditySensorDataWithPolicies

    # start the keyboard watcher
    def waitforkeypress():
        stuff = input("\n\tPress Enter to quit.\n\n")
        global EXITSAMPLE
        EXITSAMPLE = True

    threading.Thread(
        name="iotcs.sample.DirectlyConnectedDeviceSample.waitforkeypress",
        target=waitforkeypress, daemon=True).start()

    while True:
        try:
            processTemperatureSensorData(virtualTemperatureSensor,
                                         temperatureSensor,
                                         temperatureSensorState)
            processHumiditySensorData(virtualHumiditySensor, humiditySensor,
                                      humiditySensorState)
            # Wait SENSOR_POLLING_INTERVAL seconds before sending the next
            # reading.
            # The SENSOR_POLLING_INTERVAL is broken into a number of smaller
            # increments to make the application more responsive to a keypress,
            #  which will cause the application to exit.
            for i in range(NUMBEROFLOOPS):
                time.sleep(sleeptime)
            if EXITSAMPLE:
                break
        except KeyboardInterrupt:
            break


if __name__ == "__main__":
    main()
