"""iotcs.sample.MotionActivatedCamera.py."""
# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and 
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

import datetime
import logging
import os
import pathlib
import sys
import threading
import time

from iotcs.client import AbstractVirtualDevice
from iotcs.client import StorageObject
from iotcs.client.device import DirectlyConnectedDevice
from iotcs.client.device import VirtualDevice
from iotcs.common import TimeManager

# com.oracle.iot.sample

# The IOT Cloud Service device model urn for the simulated motion activated
# camera
MOTION_ACTIVATED_CAMERA_MODEL_URN = \
    "urn:com:oracle:iot:device:motion_activated_camera"
# The device model data format 
MOTION_ACTIVATED_CAMERA_MODEL_URN_DATAFORMAT = \
    f'{MOTION_ACTIVATED_CAMERA_MODEL_URN}:recording'

# Keyboard watcher flag
EXITSAMPLE = False

def getFormattedTime(dateandtime=None):
    if dateandtime is None:
        return time.strftime("%a %b %d %H:%M:%S %Z %Y")
    else:
        # its a datetime
        return dateandtime.time().strftime("%a %b %d %H:%M:%S %Z %Y")

def videoUpload(virtualDevice, dcd,
                    video, duration, startTime,
                    syncCallback):
    """
    Use a `Thread` to simulate recording.

    Use the record time to find the canned video in the
    file system, create a :class:`.StorageObject` by calling
    :meth:`.DirectlyConnectedDevice.createStorageObject`.
    Create a custom data message :class:`.Data`and
    set the storage object instance as its `video`
    attribute value, and call :meth:`.Data.submit`.

    This synchronizes the storage object with the
    storage object service.

    :param virtualdevice: the simulated camera device
    :param dcd: the DirectlyConnectedDevice
    :param duration: the time to record in seconds
    :param startTime: when to start recording
    :param syncCallback: storage object onSync callback
    """

    # Simulate the time it takes to record the video by sleeping
    try:
        time.sleep(duration)
    except Exception as e:
        print(f'MotionActivatedCameraSample.videoUpload(): '
              f'sleep interrupted, {e}')

    try:
        # The name of video will be automatically
        # prefixed with the device's client and "/"
        storageObject = dcd.createStorageObject(
            video.name, "video/mp4")
        storageObject.setOnSync(syncCallback)
        storageObject.setInputPath(video.resolve())
        storageObject.setCustomMetadata("duration", str(duration))
        consoleMessage = f'{getFormattedTime()} : ' \
                         f'{virtualDevice.getEndpointId()} : Set : ' \
                         f'"recording"={storageObject.getURI()}'

        recording = virtualDevice.createData(
            MOTION_ACTIVATED_CAMERA_MODEL_URN_DATAFORMAT)
        recording.set("video", storageObject)
        recording.set("startTime", startTime)
        recording.set("duration", duration)
        recording.submit()

        print(consoleMessage)
    except Exception as e:
        print(f'MotionActivatedCameraSample.videoUpload(): '
              f'caught exception submiting data, {e}')

def main():
    """
    This sample demonstrates the :class:`.StorageObject` and related API.
    
    The sample uses the virtualization API to simulate a motion activated
    camera.  An image is uploaded to the storage cloud every 5 seconds.
    The sample also handles the "recording" action.
    """

    #logging.basicConfig(level=logging.DEBUG)

    def showUsage():
        print(f'Usage: python MotionActivatedCameraSample.py '
              f'<provisioner file> <passphrase>')


    class CameraSyncCallback(StorageObject.SyncCallback):
        """
        A callback class to handle :meth:`.client..StorageObject.sync` events.

        This callback is set on a :class:`.iotcs.client.StorageObject`.
        It is called when the client library synchronizes a
        :class:`.client.StorageObject`. This implementation simply prints out
        the status of synchronizing the content with the Oracle Storage Cloud
        Service. The same callback is used for all of the StorageObject
        instances.
        """
        def __init__(self):
            """
            Create a :class:`.StorageObject.SyncCallback` instance.
            """
            super(CameraSyncCallback, self).__init__()

        def onSync(self, event):
            """
            Called when the library synchronizes content with the storage
            service.

            :param event: A :class:`.StorageObject.SyncEvent` instance
            """
            virtualDevice = event.getVirtualDevice()
            storageObject = event.getSource()
            msg = f'{getFormattedTime()} : '\
                  f'{virtualDevice.getEndpointId()} : onSync : '\
                  f'{storageObject.getName()}='\
                  f'"{storageObject.getSyncStatus()}"'

            if storageObject.getSyncStatus() == \
                    StorageObject.SyncStatus.IN_SYNC:
                msg = f'{msg} ({storageObject.getLength()} bytes)'

            print(msg)

    # Use this instance for all sync callbacks
    syncCallback = CameraSyncCallback()

    class SampleErrorCallback(AbstractVirtualDevice.ErrorCallback):
        def onError(self, event):
            """
            Print the event error message.

            :param event: instance of VirtualDeviceBase.ErrorEvent
            """
            virtualDevice = event.getVirtualDevice()
            msg = f'{getFormattedTime()} :'\
                  f'{virtualDevice.getEndpointId()} : onError'

            prefix = " : "

            namedvalue = event.getNamedValue()
            while namedvalue is not None:
                name = namedvalue.getName()
                value = namedvalue.getValue()
                msg = f'{msg}{prefix}"{name}"={value}'
                prefix = ","
                namedvalue = namedvalue.next()

            msg = f'{msg} : {event.getMessage()}'
            print(msg)

    class RecordActionCallback(VirtualDevice.ActionCallback):
        """
        A :class:`.ActionCallback` implementation to handle the `record` action.
        """

        __slots__ = (
            'videos',
            # a list of simulated recordings
            'dcd'
            # a :class:`.DirectlyConnectedDevice` instance
        )

        def __init__(self, dcd, videos):
            """
            Create a `RecordActionCallback` instance.

            Pass needed context to the instance.

            :param videos: a list of videos simulating camera recordings.
            """
            super(RecordActionCallback, self).__init__()
            self.videos = videos
            self.dcd = dcd

        def onAction(self, actionEvent):
            """
            Handle a request to record for some number of seconds.

            This implementation simulates recording by using a fixed set of
            timed videos, named based on the recorded time interval. The
            action event :class:`.NamedValue` argument is the amount of time
            to record. Use the value to find the video
            on the file system and call :meth:`videoUpload` in a `Thread`
            to synchronize the content with the storage service

            :param actionEvent: the action arguments
            """

            virtualDevice = actionEvent.getVirtualDevice()
            recordtime = actionEvent.getNamedValue().getValue()

            # This shouldn't happen if upper and lower bound are defined in the model.
            if recordtime is None or recordtime <= 0:
                r = 'None' if recordtime is None else recordtime
                print(f'{getFormattedTime()} : {virtualDevice.getEndpointId()}'
                      f' : onAction : Not recording "record"={r}')
                return

            d = recordtime
            # round to nearest increment of 15
            duration = int((d + 14) / 15) * 15
            # assumes videos.length > 0 and videos are 15 second increments.
            videoindex = min(int(duration/15) - 1, len(self.videos) - 1)
            video = self.videos[videoindex]

            msg = f'{getFormattedTime()} : '\
                  f'{virtualDevice.getEndpointId()} : onAction : "record"={d}'
            print(msg)

            # Start the recording now
            delay = 0
            startTime = TimeManager.currentTimeMillis()
            tt = threading.Timer(delay, videoUpload,
                args=[virtualDevice, self.dcd, video, duration,
                      startTime, syncCallback])
            tt.setName("iotcs.sample.MotionActivatedCamera.record")
            tt.setDaemon(True)
            tt.start()

    class SchedulRecordingActionCallback(VirtualDevice.ActionCallback):
        """
        A :class:`.Callable` implementation to handle the `record` action.
        """

        __slots__ = (
            'videos',
            # a list of simulated recordings
            'dcd'
            # a :class:`.DirectlyConnectedDevice` instance
        )

        def __init__(self, dcd, videos):
            """
            Create a `RecordActionCallback` instance.

            Pass needed context to the instance.

            :param videos: a list of videos simulating camera recordings.
            """
            super(SchedulRecordingActionCallback, self).__init__()
            self.videos = videos
            self.dcd = dcd

        def onAction(self, actionEvent):
            """
            Handle a request to record for `recordtime` seconds.

            This implementation simulates recording by using a fixed set of
            timed videos, named based on the recorded time interval. Use
            `recordtime` to find the video on the file system and call
            :meth:`.StorageObject.sync` to synchronize the content with
            the storage service

            :param virtualDevice: the virtualized motion activate camera device
            :param recordtime: the amount of time to record in seconds.
            """
            startTime = None
            duration = None
            virtualDevice = actionEvent.getVirtualDevice()
            argument = actionEvent.getNamedValue()
            while argument is not None:
                if argument.getName() == "startTime":
                    # arg value is in milliseconds
                    startTime = argument.getValue()
                elif argument.getName() == "duration":
                    duration = argument.getValue()
                argument = argument.next()

            if duration is None or duration <= 0:
                r = 'None' if duration is None else duration
                print(f'{getFormattedTime()} : {virtualDevice.getEndpointId()}'
                      f' : onAction : Not scheduling a recording "duration"={r}')
                return

            # round to nearest increment of 15
            duration = int((duration + 14) / 15) * 15
            # assumes videos.length > 0 and videos are 15 second increments.
            videoindex = min(int(duration/15) - 1, len(self.videos) - 1)
            video = self.videos[videoindex]
            # Convert startTime to datetime
            startTime_dt = datetime.datetime.fromtimestamp(startTime / 1000.0)
            msg = f'{getFormattedTime()} : ' \
                  f'{virtualDevice.getEndpointId()} : onAction : scheduleRecording '\
                  f'"startTime"={getFormattedTime(startTime_dt)}, "duration"={duration}'
            print(msg)

            def runCamera(virtualdevice, video):
                """
                Use a `Thread` to simulate recording.

                Use the record time to find the canned video in the
                file system, create a :class:`.StorageObject` by calling
                :meth:`.DirectlyConnectedDevice.createStorageObject`.
                Create a custom data message :class:`.Data`and
                set the storage object instance as its `video`
                attribute value, and call :meth:`.Data.submit`.

                This synchronizes the storage object with the
                storage object service.

                :param virtualdevice: the simulated camera device
                :param recordtime: the time to record in seconds
                """
                startTime = TimeManager.currentTimeMillis()

                # Simulate the time it takes to record the video by sleeping
                try:
                    time.sleep(duration)
                except Exception as e:
                    print(f'MotionActivatedCameraSample.runCamera(): '
                          f'sleep interrupted, {e}')

                try:
                    # The name of video will be automatically
                    # prefixed with the device's client and "/"
                    storageObject = self.dcd.createStorageObject(
                            video.name, "video/mp4")
                    storageObject.setOnSync(syncCallback)
                    storageObject.setInputPath(video.resolve())
                    storageObject.setCustomMetadata("duration", str(duration))
                    consoleMessage = f'{getFormattedTime()} : '\
                            f'{virtualdevice.getEndpointId()} : Set : '\
                            f'"recording"={storageObject.getURI()}'

                    recording = virtualDevice.createData(
                        MOTION_ACTIVATED_CAMERA_MODEL_URN_DATAFORMAT)
                    recording.set("video", storageObject)
                    recording.set("startTime", startTime)
                    recording.set("duration", duration)
                    recording.submit()

                    print(consoleMessage)

                except Exception as e:
                    print(f'MotionActivatedCameraSample.runCamera(): '
                        f'caught exception submiting data, {e}')

            threading.Thread(
                name="iotcs.sample.MotionActivatedCamera.runCamera",
                target=runCamera, daemon=True,
                args=[virtualDevice,video]).start()

    def getFiles(dir, ext):
        """
        Return a list of files with *ext*

        :param dir: the `Path` directory of files
        :param ext: the file extension
        :return: a list of files or `None`
        """
        files = None
        if not dir.is_dir():
            print(f'{dir} is not a directory.')
            return None

        files = []
        # produces the correct ordering
        for file in dir.iterdir():
            if file.is_file() and file.suffix == ext:
                files.append(file)

        if len(files) == 0:
            print("no files in {dir} with extension {ext}")
            return None
        files.sort()
        return files

    # TODO: make polling interval an application argument
    SLEEPTIME = 1000
    SENSORPOLLINGINTERVAL = 5000
    # default is 5 loops
    NUMBEROFLOOPS = 1 if SLEEPTIME > SENSORPOLLINGINTERVAL else \
            SENSORPOLLINGINTERVAL / SLEEPTIME
    # default ends up 1 second sleep time
    sleeptime = SENSORPOLLINGINTERVAL \
        if SLEEPTIME > SENSORPOLLINGINTERVAL else \
            SLEEPTIME + (SENSORPOLLINGINTERVAL -
                         NUMBEROFLOOPS * SLEEPTIME) / NUMBEROFLOOPS
    # make sleeptime into seconds
    sleeptime = int(sleeptime/1000)

    exiting = False

    IMAGE_ATTRIBUTE = "image"
    IMAGE_TIME_ATTRIBUTE = "imageTime"

    path = None
    passphrase = None
    no_args = len(sys.argv)

    if no_args < 3:
        showUsage()
        exit(1)
    else:
        path = sys.argv[1]
        passphrase = sys.argv[2]

    # Initialize the device client
    directlyConnectedDevice = None
    directlyConnectedDevice = DirectlyConnectedDevice(path, passphrase)

    # Activate the device
    if not directlyConnectedDevice.isActivated():
        print("Activating...")
        directlyConnectedDevice.activate([MOTION_ACTIVATED_CAMERA_MODEL_URN])

    # Create a virtual device implementing the device model
    deviceModel = directlyConnectedDevice.getDeviceModel(
        MOTION_ACTIVATED_CAMERA_MODEL_URN)

    virtualMotionActivatedCamera = \
        directlyConnectedDevice.createVirtualDevice(
            directlyConnectedDevice.getEndpointId(), deviceModel)

    # The assumption is that "images" and "videos" are in the same directory
    # as the MotionActivatedCameraSample.py file
    thisfile = os.path.abspath(__file__)
    thisdir = os.path.dirname(thisfile)
    imagesdir = pathlib.Path(thisdir).joinpath("images")
    images = getFiles(imagesdir, ".jpeg")
    videosdir = pathlib.Path(thisdir).joinpath("videos")
    videos = getFiles(videosdir, ".mp4")

    # create and set the "record" action callback
    action = RecordActionCallback(directlyConnectedDevice, videos)
    virtualMotionActivatedCamera.setOnAction(action, "record")

    # create and set the "scheduleRecording action callback
    action = SchedulRecordingActionCallback(directlyConnectedDevice, videos)
    virtualMotionActivatedCamera.setOnAction(action, "scheduleRecording")

    # create and set the error callback
    virtualMotionActivatedCamera.setOnError(SampleErrorCallback())
    print(f'\nCreated virtual motion activated camera '
          f'{virtualMotionActivatedCamera.getEndpointId()}')

    # start the keyboard watcher
    def waitforkeypress():
        stuff = input("\n\tPress Enter to quit.\n\n")
        global EXITSAMPLE
        EXITSAMPLE = True

    threading.Thread(
        name="iotcs.sample.MotionActivatedCamera.waitforkeypress",
        target=waitforkeypress, daemon=True).start()

    n = 0
    while True:
        image = images[n % len(images)]
        n += 1
        consoleMessage = f'{getFormattedTime()} : '\
            f'{virtualMotionActivatedCamera.getEndpointId()} : Set : '\
            f'"{IMAGE_ATTRIBUTE}"={image}'

        storageObject = directlyConnectedDevice.createStorageObject(
            image.name, "image/jpeg")
        storageObject.setOnSync(syncCallback)
        storageObject.setInputPath(image.resolve())
        virtualMotionActivatedCamera.update()
        virtualMotionActivatedCamera.set(IMAGE_ATTRIBUTE, storageObject)
        virtualMotionActivatedCamera.set(IMAGE_TIME_ATTRIBUTE,
                                         TimeManager.currentTimeMillis())
        virtualMotionActivatedCamera.finish()

        print(consoleMessage)
        i = 0

        # Wait SENSOR_POLLING_INTERVAL seconds before sending the next
        # image. The SENSOR_POLLING_INTERVAL is broken into a number of
        # smaller increments to make the application more responsive to a
        # keypress, which will cause the application to exit.
        try:
            for i in range(int(NUMBEROFLOOPS)):
                time.sleep(sleeptime)

            if EXITSAMPLE:
                break
        except KeyboardInterrupt:
            break

if __name__ == '__main__':
    main()

