# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.


import math
from random import randrange

from iotcs.common import TimeManager

class HumiditySensor(object):

    __slots__ = (
        'maxThreshold', # Maximum humidity threshold, writable
        'angle', # angle of rotation for simulating humidity changes
        'hmin', # The min range for values is 0% humidity
        'hmax', # Then max range for values is 100% humidity
        'hardwareId' # A unique identifier for this sensor
    )

    def __init__(self, hardwareId):
        """
        Create a HumiditySensor
        :param hardwareId: id a unique identifier for this sensor
        """
        self.hmin = 0
        self.hmax = 100
        self.maxThreshold = 80
        self.angle = 0
        self.hardwareId = hardwareId

    def getHumidity(self):
        """
         Return  the current humidity value.

         The value returned is between 0 and 100,
         representing a percent humidity.
         :return the current humidity value.
        """
        amplitude = float((float(self.maxThreshold) - float(self.hmin))/2.0)
        delta = amplitude * math.sin(math.radians(self.angle))
        humidity = 0
        if self.angle == 90:
            humidity = self.maxThreshold + 2
        else:
            humidity = round(self.hmin + amplitude + delta)
        self.angle = (self.angle + 15) % 360
        return self.clamp(humidity, self.hmin, self.hmax)

    def setMaxThreshold(self, threshold):
        """
        Set the maximum percent humidity threshold for alerts.

        The value is clamped to the range [60..100] and represents a whole
        percentage.
        :param threshold: a value between 60 and 100
        """
        if threshold < 60:
            self.maxThreshold = 60
        elif threshold > 100:
            self.maxThreshold = 100
        else:
            self.maxThreshold = threshold


    def getMaxThreshold(self):
        """
        Return the maximum threshold value.

        :return: the maxThreshold value
        """
        return self.maxThreshold

    def getManufacturer(self):
        """
        Return the manufacturer name which can be used as part of the device meta-data. # noqa

        :return: 'Sample'
        """
        return 'Sample'

    def getModelNumber(self):
        """
        Return the model number which can be used as part of the device meta-data.

        :return: the hardware id with the prefis 'MN-'
        """
        return f'MN-{self.hardwareId}'

    def getSerialNumber(self):
        """
        Return the serial number, which can be used as part of the device meta-data. # noqa

        :return: the hardware id with the prefix 'SN-'
        """
        return f'SN-{self.hardwareId}'

    def getHardwareId(self):
        """
        Return the hardware id which can be used as part of the device meta-data.

        :return: the hardware id
        """
        return self.hardwareId

    def clamp(self, val, minimum, maximum):
        """
        Clamp value between min and max, inclusive.

        :param val: the value
        :param minumum: the minimum allowable value
        :param maximum: the maximum allowable value
        :return:
        """
        return max(minimum, min(maximum, val))

class TemperatureSensor(object):
    MIN_TEMPERATURE = 10
    MAX_TEMPERATURE = 80
    # The maximum temperature this sensor can record, according to the device model.
    UPPER_TEMP_LIMIT = 80.0
    # The minimum temperature this sensor can record, according to the device model.
    LOWER_TEMP_LIMIT = -20.0

    __slots__ = (
        'angle', # the angle of rotation for simulating changing temperature
        '_power', # Whether the sensor is powered on (`True`) or off (`False`)
        'startTime', # The time (measured in EPOCH) at which the system was powered ON or RESET
        'minTemp', # The minimum value measured by the sensor since power ON or RESET
        'maxTemp', # The maximum value measured by the sensor since power ON or RESET
        'minThreshold', # The minimum threshold is in the range [-20, 0]
        'maxThreshold', #The maximum threshold is in the range [65, 80]
        'hardwareId' # The unique identifier for this sensor
    )

    def __init__(self, id):
        """
        Create a `TemperatureSensor`
        :param id: id is a unique identifier for this sensor
        """
        self.hardwareId = id
        self._power = True
        self.startTime = TimeManager.currentTimeMillis()
        self.minTemp = float('inf')
        self.maxTemp =  float('-inf')
        self.minThreshold = 0
        self.maxThreshold = 70
        self.angle = 0



    def power(self,on):
        """
        Set the power on or off.
        :param on: `True` to set power on, `False` to set power off
        """
        if on:
            if self._power:
                return
            self._power = True
            # TODO: What should this really be ? 'datetime' ? in UTC ?
            self.startTime = TimeManager.currentTimeMillis()
            self.reset()
            return

        self._power = False

    def isPoweredOn(self):
        """
        :return: `True` if powered on else `False` powered off
        """
        return self._power

    def reset(self):
        """
        An action to reset the minimum and maximum measured tempererature
        to the current value
        """
        # TODO: But this doesn't set to the current value
        self.minTemp = float('inf')
        self.maxTemp = float('-inf')

    def getTemp(self):
        amplitude = float((self.maxThreshold - self.minThreshold)/2.0)
        delta = float(amplitude * math.sin(math.radians(self.angle)))
        temp = float(0)
        if self.angle == 90:
            # violate max threshold when angle is 90 degrees
            temp = round(float(self.maxThreshold) + 2.0, 1)
        elif self.angle == 180:
            # violate min threshold when angle is 180 degrees
            temp = round(float(self.minThreshold) - 2.0, 1)
        else:
            temp = round(float(self.minThreshold) + float(amplitude) + float(delta), 1)

        self.angle = (self.angle + 15) % 360
        temp = self.clamp(temp, TemperatureSensor.LOWER_TEMP_LIMIT,
                                TemperatureSensor.UPPER_TEMP_LIMIT)
        if temp < self.minTemp:
            self.minTemp = temp
        if temp > self.maxTemp:
            self.maxTemp = temp
        return temp

    def getUnit(self):
        """
        :return: the temerature units scale
        """
        return "C" # u'\u00B0' + 'C'

    def getMinTemp(self):
        """
        Get the minimum temperature since the last power ON or RESET
        :return: the minimum temperature since the last power ON or RESET
        """
        return self.minTemp

    def getMaxTemp(self):
        """
        Get the maximum temperature since the last power ON or RESET
        :return: the maximum temperature since the last power ON or RESET
        """
        return self.maxTemp

    def getMinThreshold(self):
        """
        Get the minimum temperature threshold value
        :return: the minimum temperature threshold value
        """
        return self.minThreshold

    def setMinThreshold(self, threshold):
        """
        Set the minimum temperature threshold in degrees Celsius for alerts.
        The value is clamped at [-20, 0]
        :param threshold: the minimum temperature threshold in degrees Celsius
        :return:
        """
        if threshold < -20:
            self.minThreshold = -20
        elif threshold > 0:
            self.minThreshold = 0
        else:
            self.minThreshold = threshold

    def getMaxThreshold(self):
        """
        Get the maximum temperature threshold value
        :return: the maximum temperature threshold value
        """
        return self.maxThreshold

    def setMaxThreshold(self, threshold):
        """
        Set the maximum temperature threshold in degrees Celcius, for alerts.
        The value is clamped to the range [65, 80] inclusive
        :param threshold: the maximum temperature threshold in degrees Celcius
        """
        if threshold < 65:
            self.maxThreshold = 65
        elif threshold > 80:
            self.maxThreshold = 100
        else:
            self.maxThreshold = threshold

    def getStartTime(self):
        """
        Get the time at which the device was powered ON or RESET
        :return:
        """
        return self.startTime

    def getManufacturer(self):
        """
        Get the manufacturer name which can be used as part of the device meta-data
        :return: 'Sample'
        """
        return 'Sample'

    def getModelNumber(self):
        """
        Get the model number which can be used as part of the device meta-data
        :return: the hardware id with the prefis 'MN-'
        """
        return f'MN-{self.hardwareId}'

    def getSerialNumber(self):
        """
        Get the serial number, which can be used as part of the device meta-data
        :return: the hardware id with the prefix 'SN-'
        """
        return f'SN-{self.hardwareId}'

    def getHardwareId(self):
        """
        Get the hardware id which can be used as part of the device meta-data
        :return: the hardware id
        """
        return self.hardwareId

    def clamp(self, val, min, max):
        """
        Clamp value between min and max, inclusive.

        :param val: the value
        :param min: the minimum allowable value
        :param max: the maximum allowable value
        :return: A value such that 'min <= value <= max'
        """

        if val < min:
            return min
        if val > max:
            return max

        return val


